==> cpm-1.1.0.tar.gz: cpm-1.1.0/models/lininterp.c <==
#include "lininterp.h"
#include "mpib_utilities.h"

LinInterp_model* LinInterp_alloc(int size, int num_points) {
	LinInterp_model* model = (LinInterp_model*) malloc(sizeof(LinInterp_model));
	model->num_points = num_points;
	model->n = size;
	model->predictor.predict_p2p = LinInterp_predict_p2p;
	model->workspace = (gsl_interp**) malloc (MPIB_C2(size)*sizeof(gsl_interp *));
	model->x = (double**) malloc (MPIB_C2(size)*sizeof(double *));
	model->y = (double**) malloc (MPIB_C2(size)*sizeof(double *));
	int i;
	for (i=0; i< MPIB_C2(size) ; i++) {
		model->workspace[i] = gsl_interp_alloc(gsl_interp_linear, num_points);
		model->x[i] = (double *) malloc(num_points*sizeof(double));
		model->y[i] = (double *) malloc(num_points*sizeof(double));
	}
	//model->accel = gsl_interp_accel_alloc();
	return model;
}


void LinInterp_free(LinInterp_model* model) {
	int i;
	if (model) {
		for (i=0; i< MPIB_C2(model->n); i++) {
			gsl_interp_free(model->workspace[i]);
			free(model->x[i]);
			free(model->y[i]);
		}
		free(model->x);
		free(model->y);
		free(model->workspace);
	}
}

double LinInterp_predict_p2p(void* _this, int i, int j, int M) {
	LinInterp_model* model = (LinInterp_model*) _this;
#ifdef CPM_DEBUG
	if (i == j) {
		fprintf(stderr, "predict_p2p  with i = j not allowed!\n");
		exit(-1);
	}
#endif
	int index = MPIB_IJ2INDEX(model->n, i, j);
#ifdef CPM_DEBUG
	if (gsl_interp_eval(model->workspace[index], model->x[index], model->y[index], M, /*model->accel*/ NULL) < 0.) {
		fprintf(stderr, "Negative prediction value for message size %d and process link %d-%d\n", M, i, j);
		exit(-1);
	}
#endif
	return gsl_interp_eval(model->workspace[index], model->x[index], model->y[index], M, /*model->accel*/ NULL);
}


void LinInterp_read(FILE* stream, LinInterp_model** _model) {
	int size;
	int num_points;
	fscanf(stream, "#Number of links %d\n", &size);
	fscanf(stream, "#Number of points per link: %d\n", &num_points);
	LinInterp_model* model = *_model = LinInterp_alloc(size, num_points);
	int i, index;
	int dummy;
	for(i = 0, index = 0; i < model->n - 1; i++) {
		int j;
		for(j = i + 1; j < model->n; j++, index++) {
			int k;
			for (k = 0; k < model->num_points; k++) {
				fscanf(stream, "#%d-%d:\t%le\t%le\n", &dummy, &dummy, &model->x[index][k], &model->y[index][k]);
			}
			gsl_interp_init(model->workspace[index], model->x[index], model->y[index], model->num_points);
		}
	}
}

void LinInterp_write(FILE* stream, const LinInterp_model* model) { 
	fprintf(stream, "#Number of links %d\n", model->n);
	fprintf(stream, "#Number of points per link: %d\n", model->num_points);
	int i, index;
	for(i = 0, index = 0; i < model->n - 1; i++) {
		int j;
		for(j = i + 1; j < model->n; j++, index++) {
			int k;
			for (k = 0; k < model->num_points; k++) {
				fprintf(stream, "#%d-%d:\t%le\t%le\n", i, j, model->x[index][k], model->y[index][k]);
			}
		}
	}
}

void LinInterp_estimate(MPI_Comm comm, MPIB_precision precision, MPIB_msgset msgset, int parallel,
		 LinInterp_model** _model) {
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (size < 2) {
		if (rank == 0)
			fprintf(stderr, "Cannot compute LinInterp parameters for %d processes (must be >= 2)\n", size);
		return;
	}
	int num_points = (msgset.max_size - msgset.min_size) /msgset.stride;
        if ((msgset.max_size-msgset.min_size) % msgset.stride > 0 ) num_points++;
	LinInterp_model* model = rank == 0 ? LinInterp_alloc(size, num_points) : NULL;
	MPIB_p2p_container* container = MPIB_Send_Recv_container_alloc();

	int msg_count,peer_count;
	double* M = NULL;
	double* T = NULL;
	MPIB_pairs* pairs = MPIB_build_pairs(size);
	MPIB_pairs* iter_pairs = pairs;

	peer_count = 0;
	while (iter_pairs) {
		MPI_Barrier(comm);
		MPIB_pair* iter_pair = iter_pairs->list;
		while (iter_pair) {
			if (!parallel)
				MPI_Barrier(comm);
			if (rank == iter_pair->values[0] || rank == iter_pair->values[1]) {
				//Only process i in pair (i,j) keeps track of the benchmarks
				if (rank == iter_pair->values[0]) {
					M = (double *) realloc (M, (peer_count+1) * num_points * sizeof(double) );
					T = (double *) realloc (T, (peer_count+1) * num_points * sizeof(double) );
				}
				int m;
				msg_count = 0;
				for (m = msgset.min_size; m < msgset.max_size; m += msgset.stride) {
					MPIB_result result;
					MPIB_measure_p2p(container, comm, iter_pair->values[0], iter_pair->values[1], m, precision, &result);
					if (rank == iter_pair->values[0]) {
						M[peer_count * num_points + msg_count] = m;
						//Divide by 2 since we have a Sendrecv container
						T[peer_count * num_points + msg_count] = result.T / 2.;
					}
					msg_count++;
				}
				if (rank == iter_pair->values[0]) peer_count++;
			}
			iter_pair = iter_pair->next;
		}
		iter_pairs = iter_pairs->next;
	}
	MPIB_p2p_container_free(container);
	double* M_global;
	double* T_global;
	int *recvcounts;
 	int* displs;
	if (rank == 0) {
		recvcounts = (int*) malloc(sizeof(int)*size);
		displs = (int*)malloc(sizeof(int) * size);
		int i;
		for (i=0; i < size; i++) {
			recvcounts[i]=(size-i-1) * num_points;
			displs[i] = i == 0 ? 0 : displs[i-1] + recvcounts[i-1];
		}		
		M_global = (double*) malloc (MPIB_C2(size) * num_points * sizeof(double));
		T_global = (double*) malloc (MPIB_C2(size) * num_points * sizeof(double));
	}
	//gather the (message size,time) points at root
	MPI_Gatherv(M, peer_count * num_points, MPI_DOUBLE, rank == 0 ? M_global: NULL, recvcounts, displs, MPI_DOUBLE, 0, comm);
	MPI_Gatherv(T, peer_count * num_points, MPI_DOUBLE, rank == 0 ? T_global: NULL, recvcounts, displs, MPI_DOUBLE, 0, comm);

	int i,j;
	/* Calculate all interpolations of all link information at root process*/
	if (rank == 0 ) {
		msg_count=0;
		for (i=0; i < MPIB_C2(size) ; i++) {
			for (j=0; j<num_points;j++) {
				model->x[i][j] = M_global[msg_count];
				model->y[i][j] = T_global[msg_count++];
			}
			gsl_interp_init(model->workspace[i], model->x[i], model->y[i], model->num_points);
		}
		free(M_global);
		free(T_global);
	}
	free(M);
	free(T);
	if (rank == 0)
		*_model = model;
}

==> cpm-1.2.0.tar.gz: cpm-1.2.0/models/lininterp.c <==
#include "config.h"
#include "lininterp.h"
#include "MPIBlib/benchmarks/mpib.h"

LinInterp_model* LinInterp_alloc(int size, int num_points) {
	LinInterp_model* model = (LinInterp_model*)malloc(sizeof(LinInterp_model));
	model->predictor.predict_p2p = LinInterp_predict_p2p;
	model->n = size;
	model->num_points = num_points;
	model->splines = (gsl_spline**)malloc(MPIB_C2(size) * sizeof(gsl_spline*));
	model->accels = (gsl_interp_accel**)malloc(MPIB_C2(size) * sizeof(gsl_interp_accel*));
	int i;
	for (i = 0; i < MPIB_C2(size); i++) {
		model->splines[i] = gsl_spline_alloc(gsl_interp_linear, num_points);
		model->accels[i] = gsl_interp_accel_alloc();
	}
	return model;
}


void LinInterp_free(LinInterp_model* model) {
	int i;
	if (model) {
		for (i=0; i< MPIB_C2(model->n); i++) {
			gsl_spline_free(model->splines[i]);
			gsl_interp_accel_free(model->accels[i]);
		}
		free(model->splines);
		free(model->accels);
	}
}

double LinInterp_predict_p2p(void* _this, int i, int j, int M) {
	LinInterp_model* model = (LinInterp_model*) _this;
	if (i == j) {
		fprintf(stderr, "predict_p2p  with i = j not allowed!\n");
		exit(-1);
	}
	int index = MPIB_IJ2INDEX(model->n, i, j);
	double res = gsl_spline_eval(model->splines[index], M, model->accels[index]);
	if (res < 0.) {
		fprintf(stderr, "Negative prediction value for message size %d and process link %d-%d\n", M, i, j);
		exit(-1);
	}
	return res;
}

void LinInterp_read(FILE* stream, LinInterp_model** _model) {
	int size;
	int num_points;
	fscanf(stream, "#Number of links %d\n", &size);
	fscanf(stream, "#Number of points per link: %d\n", &num_points);
	LinInterp_model* model = *_model = LinInterp_alloc(size, num_points);
	double* x = (double*)malloc(sizeof(double) * num_points);
	double* y = (double*)malloc(sizeof(double) * num_points);
	int i, index;
	for(i = 0, index = 0; i < size - 1; i++) {
		int j;
		for(j = i + 1; j < size; j++, index++) {
			int k;
			for (k = 0; k < num_points; k++) {
				int dummy;
				fscanf(stream, "#%d-%d:\t%le\t%le\n", &dummy, &dummy, &x[k], &y[k]);
			}
			gsl_spline_init(model->splines[index], x, y, num_points);
		}
	}
	free(x);
	free(y);
}

void LinInterp_write(FILE* stream, const LinInterp_model* model) { 
	fprintf(stream, "#Number of links %d\n", model->n);
	fprintf(stream, "#Number of points per link: %d\n", model->num_points);
	int i, index;
	for(i = 0, index = 0; i < model->n - 1; i++) {
		int j;
		for(j = i + 1; j < model->n; j++, index++) {
			int k;
			for (k = 0; k < model->num_points; k++)
				fprintf(stream, "#%d-%d:\t%le\t%le\n", i, j, model->splines[index]->x[k], model->splines[index]->y[k]);
		}
	}
}

void LinInterp_estimate(MPI_Comm comm, MPIB_precision precision, MPIB_msgset msgset, int parallel,
		 LinInterp_model** _model) {
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (size < 2) {
		if (rank == 0)
			fprintf(stderr, "Cannot compute LinInterp parameters for %d processes (must be >= 2)\n", size);
		return;
	}
	int num_points = (msgset.max_size - msgset.min_size) /msgset.stride;
        if ((msgset.max_size-msgset.min_size) % msgset.stride > 0 ) num_points++;
	LinInterp_model* model = rank == 0 ? LinInterp_alloc(size, num_points) : NULL;
	MPIB_p2p_container* container = MPIB_roundtrip_container_alloc(MPI_Send, MPI_Recv);

	int msg_count,peer_count;
	double* M = NULL;
	double* T = NULL;
	MPIB_pairs* pairs = MPIB_build_pairs(size);
	MPIB_pairs* iter_pairs = pairs;

	peer_count = 0;
	while (iter_pairs) {
		MPI_Barrier(comm);
		MPIB_pair* iter_pair = iter_pairs->list;
		while (iter_pair) {
			if (!parallel)
				MPI_Barrier(comm);
			if (rank == iter_pair->values[0] || rank == iter_pair->values[1]) {
				//Only process i in pair (i,j) keeps track of the benchmarks
				if (rank == iter_pair->values[0]) {
					M = (double *) realloc (M, (peer_count+1) * num_points * sizeof(double) );
					T = (double *) realloc (T, (peer_count+1) * num_points * sizeof(double) );
				}
				int m;
				msg_count = 0;
				for (m = msgset.min_size; m < msgset.max_size; m += msgset.stride) {
					MPIB_result result;
					MPIB_measure_p2p(container, comm, iter_pair->values[0], iter_pair->values[1], m, precision, &result);
					if (rank == iter_pair->values[0]) {
						M[peer_count * num_points + msg_count] = m;
						//Divide by 2 since we have a Sendrecv container
						T[peer_count * num_points + msg_count] = result.T / 2.;
					}
					msg_count++;
				}
				if (rank == iter_pair->values[0]) peer_count++;
			}
			iter_pair = iter_pair->next;
		}
		iter_pairs = iter_pairs->next;
	}
	MPIB_roundtrip_container_free(container);
	double* M_global = NULL;
	double* T_global = NULL;
	int* recvcounts = NULL;
 	int* displs = NULL;
	if (rank == 0) {
		recvcounts = (int*) malloc(sizeof(int)*size);
		displs = (int*)malloc(sizeof(int) * size);
		int i;
		for (i=0; i < size; i++) {
			recvcounts[i] = (size - i - 1) * num_points;
			displs[i] = i == 0 ? 0 : displs[i - 1] + recvcounts[i - 1];
		}		
		M_global = (double*)malloc(MPIB_C2(size) * num_points * sizeof(double));
		T_global = (double*)malloc(MPIB_C2(size) * num_points * sizeof(double));
	}
	//gather the (message size,time) points at root
	MPI_Gatherv(M, peer_count * num_points, MPI_DOUBLE, M_global, recvcounts, displs, MPI_DOUBLE, 0, comm);
	MPI_Gatherv(T, peer_count * num_points, MPI_DOUBLE, T_global, recvcounts, displs, MPI_DOUBLE, 0, comm);
	free(recvcounts);
	free(displs);
	free(M);
	free(T);

	/* Calculate all interpolations of all link information at root process*/
	if (rank == 0 ) {
		int i;
		for (i = 0; i < MPIB_C2(size); i++)
			gsl_spline_init(model->splines[i], &M_global[i * num_points], &T_global[i * num_points], num_points);
	}
	free(M_global);
	free(T_global);
	if (rank == 0)
		*_model = model;
}
